**********************************************************************
* Author:	Han Duong
* Email:	hduong_ihatespam@ju.edu (remove _ihatespam)
* Modulename	DOROMP2
* Date:		August 31, 2010
**********************************************************************

**********************************************************************
*		DOROMP2 specification
**********************************************************************
* ASSEMBLE
*		CON(5)	=DOROMP2
*		CON(3)	libid		xxx
*		CON(3)	libcmd		yyy
* RPL
*
* Works virtually the same as DOROMP, but assumes that the object
* contained in ROMPTR xxx yyy is a compressed string: "BZ...". This
* library adds support for auto-decompression WITHOUT first copying
* the compressed string to TEMPOB (even if from a covered port), and
* auto-execution. To the user, DOROMP2 objects are virtually the
* same as DOROMP objects. In fact, all a DOROMP2 object does is it
* looks for the corresponding ROMPTR with the same libid and libcmd,
* decompresses the object found in that ROMPTR, and executes it.
*
* The config code injects the necessary code for supporting DOROMP2
* into a safe area in RAM. The injected code looks like:
*
* ASSEMBLE
*		CON(5)	romp2_spanner
*		CON(5)	romp2_skipper
*		CON(5)	=PRLG
*		GOVLNG	romp2_executor
* RPL
**********************************************************************
* Example:	Suppose we know that ROMPTR 3F0 2 contains an object
*		that has been compressed. Then we can execute the
*		compressed code via
*
*		ASSEMBLE
*			CON(5)	=DOROMP2
*			CON(3)	#3F0
*			CON(3)	#002
*		RPL
*
* NOTES:	This is merely a proof of concept. Ideally we would do
*		things like check to make sure that the pointee of the
*		DOROMP2 object exists and is a valid compressed BZ
*		string. An excercise for the reader...
**********************************************************************
xROMID	200

ASSEMBLE
ROMID#		EQU	#200
=DOROMP2	EQU	#80A0A			change this as needed
=SpanROMP2	EQU	(DOROMP2)-10
=SkipROMP2	EQU	(DOROMP2)-5
=gc1retc	EQU	#02D59
=skipromp	EQU	#03094
=xxCONFIG
RPL
::
  SIXTYFOUR TestUserFlag
  case :: "DOROMP2 disabled" FlashMsg ERRBEEP ;
  
  CODE
		AD0EX
		D0=(5)	=USEROB
		C=DAT0	A
		AD0EX
		A=PC
		?A<C	A		* PC < USEROB ?
		GOYES	+
+		GOVLNG	=PushT/FLoop
  ENDCODE
  case :: "DOROMP2 lib covered" FlashMsg ERRBEEP ;

  [#] ROMID# TOSRRP

  ( dope up our StartupProc )
  '
  ::
    ' StartupProc
    >R 'R >R
    'REVAL
    
    CODE
		CD0EX
		RSTK=C
**********************************************************************
* Insert spanromp into RAM at address given by (=DOROMP2)-10
**********************************************************************
	  	D0=(5)	=SpanROMP2
	  	A=PC
  		LC(5)	(spanromp2)-(*)
  		C=C+A	A
		DAT0=C	A
		D0=D0+	5
**********************************************************************
* Insert skipromp into RAM at address given by (=DOROMP2)-5
**********************************************************************
		LC(5)	=skipromp
		DAT0=C	A
  		D0=D0+	5
**********************************************************************
* Insert code for DOROMP at address given by (=DOROMP)
**********************************************************************
		LC(7)	#D8028FC	=PRLG and GOVLNG
		DAT0=C	7
		D0=D0+	7
		A=PC
		LC(5)	(doromp2)-(*)
		C=C+A	A
		DAT0=C	A
		C=RSTK
		CD0EX
		A=DAT0	A
		D0=D0+	5
		PC=(A)
    ENDCODE
  ;
  
  CODE
  		GOSBVL	=PopASavptr	inject insertion code into
  		D0=(5)	=FIRSTPROC	default FIRSTPROC
  		DAT0=A	A
  		GOVLNG	=GETPTRLOOP
  ENDCODE
;

ASSEMBLE
**********************************************************************
* Code for garbage collection and returning to address in C[A]
**********************************************************************
*gc1retc	RSTK=C
*		GOSBVL	=SAVPTR
*		D0=(5)	=OSAVE
*		DAT0=A	A
*		GOSBVL	=GARBAGECOL
*		D0=(5)	=OSAVE
*		A=DAT0	A
*		C=0	A
*		DAT0=C	A
*		GOSBVL	=GETPTR
*		D=D-1	A
*		RTNNC
*		D=0	A
*		GOVLNG	=DOMEMERR
**********************************************************************
* Do garbage collection and continue with doromp2 code
**********************************************************************
gcdoromp2	C=PC
		C=C+CON	A,(romp2rtn)-(*)
*		GOTO	gc1retc		4 nibbles
		GOVLNG	=gc1retc	7 nibbles
**********************************************************************
* Indirect execution of DOROMP2; aka =DOROMP2
**********************************************************************
doromp2		GOC	gcdoromp2	3 nibbles
romp2rtn	D1=D1-	5
		DAT1=A	A
		A=PC
		LC(5)	(execromp2)-(*)
		A=A+C	A
		PC=(A)			run romp2 executor
**********************************************************************
* Direct execution of ROMP2
**********************************************************************
spanromp2	D0=D0-	5		D0 -> romp2
		AD0EX
		D0=A			A[A] = ->romp2
		D0=D0+	11		D0 -> next obj in runstream
		D=D-1	A		prepare to push pointer
		GOTO	doromp2
RPL
**********************************************************************
*		BZUEval: and BZUEval
**********************************************************************
* Description:	ROMPTRs for direct execution of compressed objects
*		Objects are decompressed without copying the source
*		object to TEMPOB. The decompressed object is then
*		executed.
*
* Usage:	:: ... BZUEval: Program ;
*
*		or
*
*		:: ... ' Program BZUEval ;
*
*		where Program is declared as a NULLNAME, xNAME, etc.
*
* WARNING:	Current implementation does not check validity of the
*		compressed object. It assumes that Program is an XLIB
*		whose content is a string of the form "BZ..." that is
*		also the proper output from compression with BZ.
**********************************************************************
NULLNAME BZUEval:
::
  'R					( pop ROMPTR from runstream )

ASSEMBLE
execromp2				label for romp2 executor
RPL
  NULLNAME BZUEval
  CODE
		GOSUB	popromp
		GOSUB	getlibinfo	A[A] ->libnum; C[A] ->libaccess
		GONC	+

noliberr	LC(5)	4		Undefined XLIB Name
		GOVLNG	=GPErrjmpC		

+		GOSUB	prepbzu		set up registers for BZU
		GOSBVL	=aBZU		decompress object
		GOSUB	liboff
		GOSBVL	=GETPTR
		A=R0
		PC=(A)
**********************************************************************
libon		C=R3			configure lib; uses C[A], D0
		?C=0	A
		RTNYES
		PC=C
**********************************************************************
liboff		C=R3			unconfig lib; uses C[A], D0
		?C=0	A
		RTNYES
		P=	1
		PC=C
**********************************************************************
popromp		GOSBVL	=PopASavptr
		D1=A
		D1=D1+	5		skip prologue
		C=0	A
		C=DAT1	X
		R3=C			R3[A] = lib id
		D1=D1+	3
		C=DAT1	X		R4[A] = lib cmd
		R4=C
		RTN
**********************************************************************
getlibinfo	D1=(5)	=ROMPTAB
		C=R3
		B=C	A		B[A] = lib id
		A=DAT1	X		A[X] = number of libs
		D1=D1+	3
-		A=A-1	X		libcount--
		RTNC
		C=DAT1	X
		D1=D1+	16
		?C#B	X
		GOYES	-
		D1=D1-	16-3
		A=DAT1	A		A[A] ->libnum
		D1=D1+	5
		C=DAT1	A		C[A] ->libaccess
		R3=C			R3[A] = ->libaccess
		RTNCC
**********************************************************************
* code below missing the following checks:
*	1. ROMPTR pointed by DOROMP2 object is valid (i.e. the
*	   command number is not larger than there are commands)
*	2. The object is a compressed string that starts with 'BZ'
**********************************************************************
prepbzu		AD1EX			D1 -> libnum
		GOSUB	libon
		D1=D1+	3+5+5		skip libid, hash & msg tables
		A=DAT1	A		get link table offset
		CD1EX
		C=C+A	A
		D1=C			D1 -> link table
		D1=D1+	5+5		skip prologue, length
		C=R4			compute 5 * libcmd
		A=C	A
		A=A+A	A
		A=A+A	A
		C=C+A	A		C[A] = offset to libcmd
		AD1EX
		C=C+A	A
		D1=C
		A=DAT1	A
		C=C+A	A
		C=C+CON	A,5+5+4		skip prologue, length, 'BZ'
		D1=C			D1 -> size of compressed data
		R4=C
		A=DAT1	A
		GOSUB	liboff		in case of Insuff. Memory err
		C=A	A
		GOSBVL	=GETTEMP	allocate; error if no room
		AD0EX			save pointer to tempob
		D1=A			D1 -> hole in tempob
		GOSUB	libon		reconfig lib
		C=R4
		D0=C			D0 -> compressed data
		RTN		
  ENDCODE
;
